#!/usr/bin/env python

"""
Program data structures. There are two separate kinds of structures: those with
context, which are the values manipulated by programs, and those without
context, which are typically constant things which are stored alongside the
program but which are wrapped in context-dependent structures in the running
program.

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.

--------

The principal value structure class in this module is the Attr class which
represents attributes of objects and retains the context of each reference to an
attribute. This class models program behaviour at run-time.

The central data structure classes in this module are the following:

  * Class
  * Function
  * Module

All of the above support the Naming interface either explicitly or through
general conformance, meaning that all can be asked to provide their 'full_name'
using the method of that name.

Additionally, all of the above also support a dictionary interface in order to
access names within their defined scopes. Specific methods also exist in order
to distinguish between certain kinds of attributes:

  * Class: class_attributes, all_class_attributes, instance_attributes, all_attributes
  * Function: parameters, locals, all_locals
  * Module: module_attributes

These specific methods are useful in certain situations.

The above classes also provide an 'astnode' attribute, indicating the AST node
where each such object is defined.
"""

from micropython.program import ReplaceableContext, PlaceholderContext
from micropython.common import *
import sys

def shortrepr(obj):
    if obj is None:
        return repr(None)
    else:
        return obj.__shortrepr__()

lambda_index = 0

def new_lambda():

    "Return a new sequence number for a lambda definition."

    global lambda_index
    lambda_index += 1
    return lambda_index

# Mix-ins and abstract classes.

class Naming:

    "A mix-in providing naming conveniences."

    def full_name(self):
        if self.name is not None:
            return self.parent.full_name() + "." + self.name
        else:
            return self.parent.full_name()

class NamespaceDict:

    "A mix-in providing dictionary methods."

    def __init__(self, module=None):
        self.module = module
        self.namespace = {}
        self.globals = set()
        self.lambdas = {}           # only really useful for functions
        self.finalised = 0

        # Attributes accessed on objects, potentially narrowing their types.
        # Specific namespaces should define known names during initialisation.
        # For example, functions can define names belonging to parameters.

        # Attribute users, defining names which use attributes.

        self.attribute_users = [{}] # stack of assignments and branches
        self.attribute_user_shelves = []

        # Suspended user details plus loop details.

        self.suspended_broken_users = []     # stack of lists of user dicts
        self.suspended_continuing_users = [] # stack of lists of user dicts

        # Scope usage, indicating the origin of names.

        self.scope_usage = [{}]     # stack of scope usage
        self.scope_shelves = []

        # Define attribute usage to identify active program sections.
        # Attribute users are AST nodes defining names.

        self.all_attribute_users = set()

    # Attribute/name definition and access.

    def __delitem__(self, name):
        del self.namespace[name]

    def has_key(self, name):
        return self.namespace.has_key(name)

    def keys(self):
        return self.namespace.keys()

    def values(self):
        return self.namespace.values()

    def items(self):
        return self.namespace.items()

    def __getitem__(self, name):
        return self.namespace[name]

    def get(self, name, default=None):
        return self.namespace.get(name, default)

    # Administrative methods.

    def finalise(self, objtable):
        self.finalise_attributes()
        self.finalise_users(objtable)

    def items_for_vacuum(self):
        return self.items() + self.lambdas.items()

    def vacuum_item(self, name):
        if self.has_key(name):
            del self[name]
            return 1
        else:
            return 0

    def add_lambda(self, obj):
        attr = Attr(None, self, obj.name)
        attr.update([self.get_context_and_value(obj)], single_assignment=1)
        self.lambdas[obj.name] = attr

    # Specialised access methods.

    def get_using_node(self, name, node):

        """
        Access the given 'name' through this namespace, making use of the module
        and builtins namespaces if necessary, annotating the given 'node' with
        the scope involved.
        """

        attr, scope, full_name = self._get_with_scope(name)

        if scope is not None:
            node._scope = scope
            self.note_scope(name, scope)

        if full_name is not None and (scope != "local" or self is self.module):
            self.use_specific_attribute(full_name, name)

        return attr

    def _get_with_scope(self, name, external=0):

        """
        Find the source of the given 'name', returning the attribute object,
        scope (constant, local, global or builtins), and the full name of the
        source namespace (or None for constants).

        If the optional 'external' flag is set to a true value, only external
        (non-local) namespaces will be involved in the search.
        """

        module = self.module
        builtins = module and module.builtins or None
        importer = module and module.importer or None

        # Constants.

        if importer is not None and importer.predefined_constants.has_key(name):
            return importer.get_predefined_constant(name), "constant", None

        # Locals.

        elif not external and self.has_key(name):
            return self[name], "local", self.full_name()

        # Globals.

        elif module is not None and module is not self and module.has_key(name):
            return module[name], "global", module.full_name()

        # Builtins.

        elif builtins is not None and builtins.has_key(name):
            return builtins[name], "builtins", builtins.full_name()

        # Unknown.

        else:
            return None, None, None

    # Attribute definition methods.

    def __setitem__(self, name, value):
        self.set(name, value)

    def set(self, name, value, single_assignment=1):

        """
        A more powerful set operation, making 'name' refer to 'value' whilst
        indicating whether a 'single_assignment' (true by default) occurs in
        this operation (or whether the operation covers potentially many
        assignments in the lifetime of a program).
        """

        if value is None:
            print >>sys.stderr, "Warning: name %r in namespace %r has an unknown value (evaluated to None)." % (name, self.full_name())
            value = make_instance()

        if name in self.globals:
            self.module.set(name, value, 0)
        else:
            self._set(name, value, single_assignment)
            self.define_scope(name, "local")

    def set_module(self, name, value):

        """
        A specialised set operation, making 'name' refer to 'value' in the
        context of making a module reference available in association with
        'name' as part of the import of that module or a submodule of that
        module.
        """

        self._set(name, value, 1)

    def _set(self, name, attr_or_value, single_assignment=1):

        """
        The underlying set operation associating 'name' with the given
        'attr_or_value'.
        See: docs/assignment.txt
        """

        # Add and/or obtain the namespace entry.

        if not self.namespace.has_key(name):
            self.namespace[name] = Attr(None, self, name)

        attr = self.namespace[name]
        self._set_using_attr(attr, attr_or_value, single_assignment)

    def _set_using_attr(self, attr, attr_or_value, single_assignment=1):

        # Handle attribute assignment as well as assignment of basic objects.
        # Attempt to fix the context if not explicitly defined.

        if isinstance(attr_or_value, Attr):
            context_values = self.get_updated_context_values(attr_or_value.context_values)
        else:
            context_values = self.get_updated_context_values([self.get_context_and_value(attr_or_value)])

        attr.update(context_values, single_assignment)

    def get_context_and_value(self, value):

        "Return a context, value tuple for the given 'value'."

        # Functions have a replaceable context.

        if isinstance(value, Function):
            return (ReplaceableContext, value)

        # Classes use placeholder contexts which cannot be replaced but which
        # do not communicate useful contextual information.

        elif isinstance(value, Class):
            return (PlaceholderContext, value)

        # Other values employ themselves as the context.

        else:
            return (value, value)

    def get_updated_context_values(self, context_values):

        """
        Adapt the contexts found in the given 'context_values', returning a new
        set.
        See: docs/assignment.txt
        """

        results = set()

        for context, value in context_values:

            # Set the context of instances to themselves.

            if isinstance(value, Instance):
                results.add((value, value))
            else:
                results.add((context, value))

        return results

    def make_global(self, name):

        "Declare 'name' as a global in the current namespace."

        if not self.namespace.has_key(name):
            self.globals.add(name)
            self.define_scope(name, "global")
            return 1
        else:
            return 0

    # Attribute positioning.

    def attributes_as_list(self):

        "Return the attributes in a list."

        self.finalise_attributes()
        l = [None] * len(self.keys())
        for attr in self.values():
            l[attr.position] = attr
        return l

    def finalise_attributes(self):

        "Make sure all attributes are fully defined."

        if self.finalised:
            return

        # The default action is to assign attribute positions sequentially.

        for i, attr in enumerate(self.values()):
            attr.position = i

        self.finalised = 1

    def unfinalise_attributes(self):

        "Open attribute definitions to editing and subsequent finalisation."

        self.finalised = 0

    # Attribute usage methods.

    def finalise_attribute_usage(self):

        "Propagate attribute usage for the namespace to the importer."

        module = self.module
        importer = module and module.importer

        if importer is not None:

            # Visit each user and examine the attribute usage for each name.

            for user in self.all_attribute_users:

                # First, visit the contributors and combine their attribute
                # usage with the usage recorded directly on the user.

                self.get_usage_from_contributors(user)

                # Record the defining user on each contributor.

                for contributor in user._attrcontributors:
                    contributor._attrdefs.append(user)

                # Then, tell the importer about the usage.

                for name in user._attrnames.keys():

                    # Only provide information about names defined by this user.

                    usage = user._attrcombined.get(name, [])

                    # Skip reporting where no actual usage occurs.

                    if usage is None:
                        continue

                    # Eliminate non-usage.

                    importer.use_names(user, name, tuple([attrnames for attrnames in usage if attrnames]), self.full_name())

    def finalise_users(self, objtable):

        "Record the object types for generating guards."

        # Visit each user and examine the attribute usage for each name.

        for user in self.all_attribute_users:
            user._attrtypes = self._deduce_types(user._attrcombined, objtable)
            self._finalise_contributor(user, objtable)

    def _finalise_contributors(self, node, objtable):

        """
        Visit the contributing branches of 'node', finalising them using the
        given 'objtable'.
        """

        for contributor in node._attrbranches:
            self._finalise_contributor(contributor, objtable)

    def _finalise_contributor(self, node, objtable):

        """
        Record the specific object types being used in various regions of a
        program unit.
        """

        if not hasattr(node, "_attrspecifictypes"):
            merged = {}

            # Get the combined usage information from the user definitions.

            for user in node._attrdefs or [node]:

                # Filter the usage for each name using the local usage
                # information.

                for name, usage in user._attrcombined.items():
                    localusage = node._attrnames.get(name)

                    if usage and localusage:
                        if not merged.has_key(name):
                            merged[name] = ObjectSet()

                        for attrnames, value in usage.items():
                            if attrnames and localusage.issubset(attrnames):
                                merged[name][attrnames] = value

            node._attrmerged = merged
            node._attrspecifictypes = self._deduce_types(node._attrmerged, objtable)

            self._finalise_contributors(node, objtable)

    def _deduce_types(self, usage, objtable):

        """
        Deduce the types for names from the given attribute 'usage' and using
        the given 'objtable'.
        """

        attrtypes = {}
        for name, combined_usage in usage.items():
            if combined_usage is not None:
                objtypes = get_object_types_for_usage(combined_usage, objtable, name, self.full_name(), True)
                if objtypes:
                    if isinstance(self, Function) and self.is_method() and name == "self":
                        objtypes = filter_using_self(objtypes, self.parent)
                attrtypes[name] = objtypes
        return attrtypes

    def get_usage_from_contributors(self, node):

        """
        Obtain usage information from the given 'node', combined with usage
        details from its contributors, returning a tuple containing a set of all
        contributors employed along with a dictionary mapping names to lists of
        usage possibilities (each a collection of attribute names).
        """

        unfinished = {}

        if not hasattr(node, "_attrcombined"):
            node._attrcombined = None
            node._attrcontributors = None

            for contributor in node._attrbranches:

                # Get contributor details.

                unfinished_contributors = self.get_usage_from_contributors(contributor)

                # Collect unfinished contributors and affected nodes.

                # Where the contributor is already set to None, a loop has
                # occurred and this node will need to have its usage
                # recalculated later for the unfinished contributor.

                if contributor._attrcombined is None:
                    if not unfinished.has_key(contributor):
                        unfinished[contributor] = []
                    unfinished[contributor].append(node)
                    continue

                # Where the contributor provides usage details, it may also
                # communicate unfinished contributor information. As a
                # consequence, this node is also affected.

                for unfinished_contributor, nodes in unfinished_contributors.items():
                    if not unfinished.has_key(unfinished_contributor):
                        unfinished[unfinished_contributor] = nodes
                    else:
                        unfinished[unfinished_contributor] += nodes

                    if node not in unfinished[unfinished_contributor]:
                        unfinished[unfinished_contributor].append(node)

            # Set the current state of the usage on this node.

            node._attrcombined, node._attrcontributors = \
                self.get_usage_from_contributors_for_node(node)

            # Complete unfinished contributors relying on this node.

            if unfinished.has_key(node):
                processed = set()
                for contributor in unfinished[node]:
                    if not contributor in processed:
                        processed.add(contributor)

                        contributor._attrcombined, contributor._attrcontributors = \
                            self.get_usage_from_contributors_for_node(contributor)

                del unfinished[node]

        return unfinished

    def get_usage_from_contributors_for_node(self, node):

        # Visit each contributor, gathering usage for each name.

        contributor_usage = {}
        all_contributions = []
        all_contributors = set()

        for contributor in node._attrbranches:
            usage = contributor._attrcombined
            if usage is not None:
                all_contributions.append(usage)

            all_contributors.add(contributor)
            contributors = contributor._attrcontributors
            if contributors is not None:
                all_contributors.update(contributors)

        # Get contributed usage for each contributor.
        # This gathers usage for each name such as {(a, b), (c, d)} and
        # {(a, b), (e, f)} into a single set {(a, b), (c, d), (e, f)}.

        update_mapping_dict(contributor_usage, all_contributions)

        # Then get the resulting usage.
        # First, make the current usage compatible with the contributed
        # usage: this makes the attribute usage for each name merely one
        # member in a list of many possibilities.
        # Then, combine the current usage with the contributed usage.
        # Thus, usage of {(f, g)} combined with {(a, b), (c, d)} would give
        # {(f, g, a, b), (f, g, c, d)}.

        return combine_mapping_dicts(deepen_mapping_dict(node._attrnames), contributor_usage), all_contributors

    def use_attribute(self, name, attrname, value=None):

        """
        Note usage on the attribute user 'name' of the attribute 'attrname',
        noting an assignment if 'value' is specified.
        """

        return self._use_attribute(name, attrname, value)

    def use_specific_attribute(self, objname, attrname):

        "Declare the usage on 'objname' of the given 'attrname'."

        self._use_specific_attribute(objname, attrname)

    # These shadow various methods in the InspectedModule class, and provide
    # implementations generally.

    def _use_specific_attribute(self, objname, attrname, from_name=None):

        """
        Note attribute usage specifically on 'objname' - an object which is
        known at inspection time - or in the current unit if 'objname' is None,
        nominating a specific attribute 'attrname'.

        This bypasses attribute user mechanisms.
        """

        from_name = from_name or self.full_name()
        objname = objname or from_name
        module = self.module
        importer = module and module.importer

        if importer is not None:
            importer.use_specific_name(objname, attrname, from_name)

    def _use_attribute(self, name, attrname, value=None):

        """
        Indicate the use of the given 'name' in this namespace of an attribute
        with the given 'attrname'. If the optional 'value' is specified, an
        assignment using the given 'value' is recorded.
        """

        users = self.attribute_users[-1]

        # Add the usage to all current users.

        if users.has_key(name):
            for user in users[name]:
                values = user._attrnames[name]
                if values is None:
                    values = user._attrnames[name] = ObjectSet()

                # Add an entry for the attribute, optionally with an assigned
                # value.

                values.add(attrname)
                if value is not None:
                    values[attrname].add(value)

            return users[name]
        else:
            return []

    def _define_attribute_user(self, node):

        """
        Define 'node' as the user of attributes, indicating the point where the
        user is defined.
        """

        name = node.name
        self._define_attribute_user_for_name(node, name)

    def _define_attribute_user_for_name(self, node, name):

        "Define 'node' as the user of attributes for the given 'name'."

        users = self.attribute_users[-1]

        # This node overrides previous definitions.

        users[name] = set([node])

        # Record the attribute combinations for the name.

        self._init_attribute_user_for_name(node, name)

        # Remember this user.

        self.all_attribute_users.add(node)

    def _init_attribute_user_for_name(self, node, name):

        "Make sure that 'node' is initialised for 'name'."

        self._init_attribute_user(node)
        node._attrnames[name] = None

    def _init_attribute_user(self, node):

        # Attribute usage for names.

        if not hasattr(node, "_attrnames"):
            node._attrnames = {}
            node._attrmerged = {}

        # Branches contributing usage to this node.

        if not hasattr(node, "_attrbranches"):
            node._attrbranches = []

        # Definitions receiving usage from this node.

        if not hasattr(node, "_attrdefs"):
            node._attrdefs = []

    # Branch management methods.

    def _new_branchpoint(self, loop_node=None):

        """
        Establish a new branchpoint where several control-flow branches diverge
        and subsequently converge.
        """

        self.attribute_user_shelves.append([])
        self.scope_shelves.append([])

        if loop_node is not None:
            self.suspended_broken_users.append([])
            self.suspended_continuing_users.append((loop_node, []))

    def _new_branch(self, node):

        """
        Establish a new control-flow branch, transferring attribute usage to
        the new branch so that it may be augmented for each name locally.

        Add the given 'node' as an active user to be informed of attribute
        usage.
        """

        attribute_users = self.attribute_users[-1]

        # Define this node as the active attribute user for all currently
        # defined names.

        new_users = {}

        for name in attribute_users.keys():
            new_users[name] = [node]
            self._init_attribute_user_for_name(node, name)

        self._init_attribute_user(node)
        self.attribute_users.append(new_users)

        # Add this user as a contributor to the previously active users.

        self._connect_users_to_branch(attribute_users, node)

        # Retain a record of scope usage.

        scope_usage = {}
        scope_usage.update(self.scope_usage[-1])
        self.scope_usage.append(scope_usage)

    def _connect_users_to_branch(self, attribute_users, node):

        """
        Given the 'attribute_users' mapping, connect the users referenced in the
        mapping to the given branch 'node'.
        """

        all_users = set()

        for users in attribute_users.values():
            all_users.update(users)

        for user in all_users:
            self._init_attribute_user(user)
            user._attrbranches.append(node)

    def _abandon_branch(self):

        """
        Abandon scope usage, permitting locally different scopes for names,
        provided these cannot "escape" from the branch.
        """

        self.attribute_users[-1] = abandoned_branch_users
        self.scope_usage[-1] = abandoned_branch_scope

    def _suspend_broken_branch(self):

        """
        Suspend a branch for resumption after the current loop.
        """

        attribute_users = self.attribute_users[-1]

        if not isinstance(attribute_users, AbandonedBranchUsers):
            users = self.suspended_broken_users[-1]
            users.append(attribute_users)
            self._abandon_branch()

    def _suspend_continuing_branch(self):

        """
        Suspend a branch for resumption after the current iteration.
        """

        attribute_users = self.attribute_users[-1]

        if not isinstance(attribute_users, AbandonedBranchUsers):
            loop_node, users = self.suspended_continuing_users[-1]
            users.append(attribute_users)
            self._abandon_branch()

    def _shelve_branch(self):

        """
        Shelve the current control-flow branch, recording the attribute usage
        for subsequent merging. If this branch should be abandoned, the usage
        observations are still recorded but will not contribute to subsequent
        observations after a merge.
        """

        users = self.attribute_users.pop()
        self.attribute_user_shelves[-1].append(users)

        scope_usage = self.scope_usage.pop()
        self.scope_shelves[-1].append(scope_usage)

    def _merge_branches(self):

        """
        Merge control-flow branches. This should find the users active within
        each branch, which have been "shelved", and update the active users
        dictionary with these contributions.
        """

        # Combine the attribute users. This ensures that a list of users
        # affected by attribute usage is maintained for the current branch.

        all_shelved_users = self.attribute_user_shelves.pop()
        new_users = merge_mapping_dicts(all_shelved_users)
        self.attribute_users[-1] = new_users

        # Combine the scope usage.

        scope_usage = self.scope_usage[-1]
        new_scope_usage = {}

        all_scope_usage = self.scope_shelves.pop()
        all_scope_names = set()

        # Find all the names for whom scope information has been defined.

        for shelved_usage in all_scope_usage:
            all_scope_names.update(shelved_usage.keys())

        for shelved_usage in all_scope_usage:
            for name in all_scope_names:

                # Find the recorded scope for the name.

                if shelved_usage.has_key(name):
                    scope = shelved_usage[name]
                elif scope_usage.has_key(name):
                    scope = scope_usage[name]

                # For abandoned branches, no scope is asserted for a name.

                elif isinstance(shelved_usage, AbandonedBranchScope):
                    scope = None

                # If no scope is recorded, find a suitable external source.

                else:
                    attr, scope, full_name = self._get_with_scope(name, external=1)

                # Attempt to record the scope, testing for conflicts.

                if scope:
                    if not new_scope_usage.has_key(name):
                        new_scope_usage[name] = scope
                    elif new_scope_usage[name] != scope:
                        new_scope_usage[name] = ScopeConflict(scope, new_scope_usage[name])

        self.scope_usage[-1] = new_scope_usage

    def _resume_broken_branches(self):

        """
        Incorporate users from suspended broken branches into the current set of
        active users.
        """

        suspended_users = self.suspended_broken_users.pop()
        current_users = self.attribute_users[-1]
        new_users = merge_mapping_dicts(suspended_users + [current_users])
        self.attribute_users[-1] = new_users

    def _resume_continuing_branches(self):

        """
        Incorporate users from suspended continuing branches into the current
        set of active users, merging usage from the latter with the former.
        """

        loop_node, suspended_users = self.suspended_continuing_users.pop()
        current_users = self.attribute_users[-1]

        # Connect the suspended users to the loop node.

        for users in suspended_users:
            self._connect_users_to_branch(users, loop_node)

        # Merge suspended branches with the current branch.

        new_users = merge_mapping_dicts(suspended_users + [current_users])
        self.attribute_users[-1] = new_users

    # Scope usage methods.

    def define_scope(self, name, scope):

        """
        Define 'name' as being from the given 'scope' in the current namespace.
        """

        self.scope_usage[-1][name] = scope

    def note_scope(self, name, scope):

        """
        Note usage of 'name' from the given 'scope' in the current namespace.
        If a conflict has been recorded previously, raise an exception.
        """

        scope_usage = self.scope_usage[-1]

        if scope_usage.has_key(name):
            found_scope = scope_usage[name]
            if isinstance(found_scope, ScopeConflict):
                raise InspectError("Scope conflict for %r: defined as both %s and %s." % (
                    name, found_scope.old_scope, found_scope.new_scope))

        scope_usage[name] = scope

    def used_in_scope(self, name, scope):

        """
        Return whether 'name' is used from the given 'scope' in the current
        namespace.
        """

        scope_usage = self.scope_usage[-1]
        return scope_usage.get(name) == scope

# Special helper classes for usage and scope resolution.

class EmptyDict:

    "A class providing dictionaries which retain no information."

    def has_key(self, name):
        return 0

    def __setitem__(self, name, value):
        pass

    def __getitem__(self, name):
        raise KeyError, name

    def get(self, name, default=None):
        return default

    def keys(self):
        return []

    values = items = keys

class AbandonedBranchUsers(EmptyDict):

    """
    A class providing a value or state for an abandoned branch distinct from an
    empty usage dictionary.
    """

    pass

class AbandonedBranchScope(EmptyDict):

    """
    A class providing a value or state for an abandoned branch distinct from an
    empty scope dictionary.
    """

    pass

abandoned_branch_users = AbandonedBranchUsers()
abandoned_branch_scope = AbandonedBranchScope()

class ScopeConflict:

    """
    A scope conflict caused when different code branches contribute different
    sources of names.
    """

    def __init__(self, old_scope, new_scope):
        self.old_scope = old_scope
        self.new_scope = new_scope

class NullBranch:

    "A class representing an attribute user for a non-existent branch."

    pass

# Program data structures.

class Attr:

    "An attribute entry having a context."

    def __init__(self, position, parent, name):

        """
        Initialise the attribute with the given 'position' within the collection
        of attributes of its 'parent', indicating its 'name'.
        """

        self.position = position
        self.parent = parent
        self.name = name

        # Possible values.

        self.context_values = set()

        # Number of assignments per name.

        self.assignments = None

    # Value-related methods.

    def get_contexts(self):
        return [c for (c, v) in self.context_values]

    def get_values(self):
        return [v for (c, v) in self.context_values]

    def get_context(self):
        if len(self.context_values) == 1:
            return self.get_contexts()[0]
        else:
            return None

    def get_value(self):
        if len(self.context_values) == 1:
            return self.get_values()[0]
        else:
            return None

    __call__ = get_value    # convenient access to any single value

    def update(self, context_values, single_assignment):

        """
        Update the attribute, adding the 'context_values' provided to the
        known details associated with the attribute, changing the number of
        assignments according to the 'single_assignment' status of the
        operation, where a true value indicates that only one assignment is
        associated with the update, and a false value indicates that potentially
        many assignments may be involved.
        """

        if self.context_values.issuperset(context_values) and \
            not (make_instance(), make_instance()) in context_values:
            return

        if self.assignments is None:
            if single_assignment:
                self.assignments = len(set(context_values))
            else:
                self.assignments = AtLeast(len(set(context_values)))
        else:
            if single_assignment:
                self.assignments += 1
            else:
                self.assignments += AtLeast(1)

        self.context_values.update(context_values)

    def is_constant(self):

        """
        Return whether this attribute references something that can be regarded
        as being constant within a particular scope.
        """

        return self.assignments == 1

    def is_strict_constant(self):

        """
        Return whether this attribute references something that can be regarded
        as being constant.
        """

        value = self.get_value()
        return not (value is None or (isinstance(value, Instance) and not isinstance(value, Constant)))

    def is_static_attribute(self):

        """
        Return whether this attribute is defined on a fixed/static object such
        as a class or a module.
        """

        return isinstance(self.parent, (Class, Module))

    def defines_ambiguous_class(self):

        "Return whether this attribute defines more than one class."

        if self.assignments > 1:
            have_class = 0
            for obj in self.get_values():
                if isinstance(obj, Class):
                    if have_class:
                        return 1
                    have_class = 1

        return 0

    def defined_within_hierarchy(self):

        """
        Return whether the parent and context of the attribute belong to the
        same class hierarchy.
        """

        # Must be defined within a class.

        if isinstance(self.parent, Class):

            # To be sure, all contexts must be classes and be the same as the
            # parent, or be a superclass of the parent, or be a subclass of the
            # parent.

            for context in self.get_contexts():
                if not (
                    isinstance(context, Class) and (
                    context is self.parent or
                    context.has_subclass(self.parent) or
                    self.parent.has_subclass(context))
                    ):
                    return 0

            return 1

        # Instance attributes are not defined within a hierarchy.

        else:
            return 0

    def defined_outside_hierarchy(self):

        """
        Return whether the parent and context of the attribute never belong to
        the same class hierarchy.
        """

        # Must be defined within a class.

        if isinstance(self.parent, Class):

            # To be sure, all contexts must be classes and be the same as the
            # parent, or be a superclass of the parent, or be a subclass of the
            # parent.

            for context in self.get_contexts():
                if not (
                    isinstance(context, Class) and not (
                    context is self.parent or
                    context.has_subclass(self.parent) or
                    self.parent.has_subclass(context))
                    ):
                    return 0

            return 1

        # Instance attributes are not defined within a hierarchy.

        else:
            return 0

    def __repr__(self):
        if self.position is not None:
            position = "at %r, " % self.position
        else:
            position = ""
        return "<attribute %s.%s (%sassigned %r)>" % (
            shortrepr(self.parent), self.name,
            position, self.assignments
            )

    def __shortrepr__(self):
        return "%s.%s (at %r)" % (shortrepr(self.parent), self.name, self.position)

    def _context_values_str(self):
        l = []
        for (c, v) in self.context_values:
            l.append("(c=%s, v=%s)" % (shortrepr(c), shortrepr(v)))
        return ", ".join(l)

# Instances are special in that they need to be wrapped together with context in
# a running program, but they are not generally constant.

class Instance:

    "A placeholder indicating the involvement of an instance."

    def __init__(self):
        self.parent = None

        # Image generation details.

        self.location = None

    def __repr__(self):
        return "<instance>"

    def __eq__(self, other):
        return other.__class__ is Instance

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return 0

    __shortrepr__ = __repr__

common_instance = Instance()

def make_instance():
    return common_instance

class Constant:

    "A superclass for all constant or context-free structures."

    pass

# Data objects appearing in programs before run-time.

class Const(Constant, Instance):

    "A constant object with no context."

    def __init__(self, value):
        Instance.__init__(self)
        self.value = value

    def get_value(self):
        return self.value

    def __repr__(self):
        if self.location is not None:
            return "Const(%r, location=%r)" % (self.value, self.location)
        else:
            return "Const(%r)" % self.value

    __shortrepr__ = __repr__

    # Support constants as dictionary keys in order to build constant tables.

    def __eq__(self, other):
        return other is not None and isinstance(other, Const) and \
            self.value == other.value and self.value.__class__ is other.value.__class__

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.value)

    # Constants are instances of various built-in types.

    def value_type_name(self):
        return ".".join(self.value_type_name_parts())

    def value_type_name_parts(self):
        return "__builtins__", self.value.__class__.__name__

class Class(NamespaceDict, Naming, Constant):

    "A base class for common/normal classes and the type class."

    def __init__(self, name, parent=None, module=None, node=None):

        """
        Initialise the class with the given 'name', optional 'parent' object,
        'module' and AST 'node'. The optional information must be set at a later
        point using the 'set_context' method if omitted.
        """

        NamespaceDict.__init__(self, module)
        self.name = name
        self.parent = parent
        self.astnode = node

        # Superclasses, descendants and attributes.

        self.bases = []
        self.descendants = set()
        self.instattr = set()           # instance attributes
        self.instattr_tentative = set() # tentative/suspected instance attributes
        self.relocated = set()          # attributes which do not have the same
                                        # position as those of the same name in
                                        # some superclasses

        # Caches.

        self.reset_caches()

        # Image generation details.

        self.location = None
        self.code_location = None
        self.code_body_location = None  # corresponds to the instantiator

        self.instantiator = None
        self.instance_template_location = None # for creating instances at run-time

        # Program-related details.

        self.blocks = None
        self.temp_usage = 0
        self.local_usage = 0
        self.all_local_usage = 0

        # Add an attribute to this class for use by instances.

        self.set("__class__", self)

    def set_context(self, parent, module, node):

        "Set the 'parent', 'module' and 'node' of a class created in advance."

        self.parent = parent
        self.module = module
        self.astnode = node

    def reset_caches(self):

        "Reset the caches."

        self.all_instattr = None        # cache for instance_attributes
        self.all_instattr_names = None  # from all_instattr
        self.all_classattr = None       # cache for all_class_attributes
        self.all_classattr_names = None # from all_classattr
        self.allattr = None             # cache for all_attributes
        self.allattr_names = None       # from allattr

    def __repr__(self):
        if self.location is not None:
            return "<class %s (at %r)>" % (shortrepr(self), self.location)
        else:
            return "<class %s>" % shortrepr(self)

    def __shortrepr__(self):
        return "%s.%s" % (shortrepr(self.parent), self.name)

    def get_body_block(self):
        return self.get_instantiator().blocks[0]

    # Namespace-related methods.

    def get_updated_context_values(self, context_values):

        """
        Adapt the contexts found in the given 'context_values', returning a new
        set.
        See: docs/assignment.txt
        """

        results = set()

        for context, value in context_values:

            # Change the ownership of functions.

            if context is ReplaceableContext and value is not None and isinstance(value, Function):
                results.add((self, value))
            else:
                results.add((context, value))

        return NamespaceDict.get_updated_context_values(self, results)

    # Administrative methods.

    def items_for_vacuum(self):

        "Consider both class and instance attributes for vacuuming."

        items = []
        for name in self.instattr:
            items.append((name, None))
        return NamespaceDict.items_for_vacuum(self) + items

    def vacuum_item(self, name):

        "Vacuum 'name' from the class or instance attribute collections."

        # NOTE: Hack to prevent damage to exceptions.

        if name == "_pc":
            return 0

        if not NamespaceDict.vacuum_item(self, name):
            self.instattr.remove(name)
        return 1

    def finalise_attributes(self):

        "Make sure that all attributes are fully defined."

        if self.finalised:
            return

        self.finalise_class_attributes()
        self.finalise_instance_attributes()
        self.finalised = 1

    def unfinalise_attributes(self):

        "Open attribute definitions to editing and subsequent finalisation."

        self.reset_caches()
        self.finalised = 0

    # Convenience methods for accessing functions and methods.

    def get_instantiator(self):

        "Return a function which can be used to instantiate the class."

        if self.instantiator is None:
            self.instantiator = self.get_init_method().as_instantiator()
        return self.instantiator

    def get_init_method(self):
        return self.all_class_attributes()["__init__"].get_value()

    # Class-specific methods.

    def add_base(self, base):
        self.bases.append(base)
        base.add_descendant(self)

    def add_instance_attribute(self, name, tentative=False):
        if tentative:
            self.instattr_tentative.add(name)
        else:
            self.instattr.add(name)

    def add_descendant(self, cls):
        self.descendants.add(cls)
        for base in self.bases:
            base.add_descendant(cls)

    def has_subclass(self, other):
        return other in self.descendants

    def all_descendants(self):
        d = {}
        for cls in self.descendants:
            d[cls.full_name()] = cls
        return d

    "Return the attribute names provided by this class only."

    class_attribute_names = NamespaceDict.keys

    def class_attributes(self):

        "Return class attributes provided by this class only."

        return dict(self)

    def all_class_attribute_names(self):

        "Return the attribute names provided by classes in this hierarchy."

        if self.all_classattr_names is None:
            self.all_class_attributes()
            self.all_classattr_names = self.all_classattr.keys()
        return self.all_classattr_names

    def all_class_attributes(self):

        "Return all class attributes, indicating the class which provides them."

        self.finalise_class_attributes()
        return self.all_classattr

    def finalise_class_attributes(self):

        "Make sure that the class attributes are fully defined."

        if self.all_classattr is None:
            self.all_classattr = {}
            clsattr = {}

            # Record provisional position information for attributes of this
            # class.

            for name in self.class_attributes().keys():

                # Special case: __class__ has to be at position 0.

                if name == "__class__":
                    clsattr[name] = set([0])
                else:
                    clsattr[name] = set() # position not yet defined

            reversed_bases = self.bases[:]
            reversed_bases.reverse()

            # For the bases in reverse order, acquire class attribute details.

            for cls in reversed_bases:
                for name, attr in cls.all_class_attributes().items():
                    self.all_classattr[name] = attr

                    # Record previous attribute information.

                    if clsattr.has_key(name):
                        clsattr[name].add(attr.position)

            # Record class attributes provided by this class and its bases,
            # along with their positions.

            self.all_classattr.update(self.class_attributes())

            if clsattr:
                for i, name in enumerate(self._get_position_list(clsattr)):
                    self.all_classattr[name].position = i

        return self.all_classattr

    def instance_attribute_names(self):

        "Return the instance attribute names provided by the class."

        if self.all_instattr_names is None:
            self.instance_attributes()
        return self.all_instattr_names

    def instance_attributes(self):

        "Return instance-only attributes for instances of this class."

        self.finalise_instance_attributes()
        return self.all_instattr

    def finalise_instance_attributes(self):

        "Make sure that the instance attributes are fully defined."

        # Eliminate tentative instance attributes that are actually class
        # attributes.

        for attrname in self.all_class_attributes().keys():
            if attrname in self.instattr_tentative:
                self.instattr_tentative.remove(attrname)

        for cls in self.descendants:
            for attrname in cls.class_attribute_names():
                if attrname in self.instattr_tentative:
                    self.instattr_tentative.remove(attrname)

        for attrname in self.instattr_tentative:
            self.instattr.add(attrname)

        # Cache the attributes by converting the positioned attributes into a
        # dictionary.

        if self.all_instattr is None:
            self.all_instattr = self._get_attributes()
            self.all_instattr_names = self.all_instattr.keys()

        return self.all_instattr

    def _get_attributes(self):

        """
        Return a dictionary mapping names to Attr instances incorporating
        information about their positions in the final instance structure.
        """

        instattr = {}

        # Record provisional position information for attributes of this
        # instance.

        for name in self.instattr:
            instattr[name] = set() # position not yet defined

        reversed_bases = self.bases[:]
        reversed_bases.reverse()

        # For the bases in reverse order, acquire instance attribute
        # details.

        for cls in reversed_bases:
            for name, attr in cls.instance_attributes().items():

                # Record previous attribute information.

                if instattr.has_key(name):
                    instattr[name].add(attr.position)
                else:
                    instattr[name] = set([attr.position])

        # Build the dictionary of attributes using the existing positions known
        # for each name.

        d = {}
        for i, name in enumerate(self._get_position_list(instattr)):
            d[name] = Attr(i, make_instance(), name)
        return d

    def _get_position_list(self, positions):

        """
        Return a list of attribute names for the given 'positions' mapping from
        names to positions, indicating the positions of the attributes in the
        final instance structure.
        """

        position_items = positions.items()
        namearray = [None] * len(position_items)

        # Get the positions in ascending order of list size, with lists
        # of the same size ordered according to their smallest position
        # value.

        position_items.sort(self._cmp_positions)

        # Get the names in position order.

        held = []

        for name, pos in position_items:
            pos = list(pos)
            pos.sort()
            if pos and pos[0] < len(namearray) and namearray[pos[0]] is None:
                namearray[pos[0]] = name
            else:
                if pos:
                    self.relocated.add(name)
                held.append((name, pos))

        for i, attr in enumerate(namearray):
            if attr is None:
                name, pos = held.pop()
                namearray[i] = name

        return namearray

    def _cmp_positions(self, a, b):

        "Compare name plus position list operands 'a' and 'b'."

        name_a, list_a = a
        name_b, list_b = b
        if len(list_a) < len(list_b):
            return -1
        elif len(list_a) > len(list_b):
            return 1
        elif not list_a:
            return 0
        else:
            return cmp(min(list_a), min(list_b))

    def all_attribute_names(self):

        """
        Return the names of all attributes provided by instances of this class.
        """

        self.allattr_names = self.allattr_names or self.all_attributes().keys()
        return self.allattr_names

    def all_attributes(self):

        """
        Return all attributes for an instance, indicating either the class which
        provides them or that the instance itself provides them.

        Note that __class__ acts like a class attribute for both instances and
        classes, and must be able to convey distinct values.
        """

        if self.allattr is None:
            self.allattr = {}
            self.allattr.update(self.all_class_attributes())
            for name, attr in self.instance_attributes().items():
                if self.allattr.has_key(name) and name != "__class__":
                    print >>sys.stderr, "Warning: instance attribute %r in %r overrides class attribute." % (name, self)
                self.allattr[name] = attr
        return self.allattr

class TypeClass(Class):

    "A special class for the type class."

    pass

class CommonClass(Class):

    "An inspected class."

    pass

class Function(NamespaceDict, Naming, Constant):

    "An inspected function."

    def __init__(self, name, parent, argnames, defaults, has_star, has_dstar,
        dynamic_def=0, module=None, node=None):

        """
        Initialise the function with the given 'name', 'parent', list of
        'argnames', list of 'defaults', the 'has_star' flag (indicating the
        presence of a * parameter), the 'has_dstar' flag (indicating the
        presence of a ** parameter), optional 'dynamic_def' (indicating that the
        function must be handled dynamically), optional 'module', and optional
        AST 'node'.
        """

        NamespaceDict.__init__(self, module)

        if name is None:
            self.name = "lambda#%d" % new_lambda()
            self._is_lambda = 1
        else:
            self.name = name
            self._is_lambda = 0

        self.parent = parent
        self.argnames = argnames
        self.defaults = defaults
        self.has_star = has_star
        self.has_dstar = has_dstar
        self.dynamic_def = dynamic_def
        self.astnode = node

        # Initialise the positional names.

        self.positional_names = self.argnames[:]
        if has_dstar:
            self.dstar_name = self.positional_names[-1]
            del self.positional_names[-1]
        if has_star:
            self.star_name = self.positional_names[-1]
            del self.positional_names[-1]

        # Initialise default storage.
        # NOTE: This must be initialised separately due to the reliance on node
        # NOTE: visiting.

        self.default_attrs = []

        # Initialise attribute usage.

        if node is not None:
            for arg in argnames:

                # Define attribute users.

                self._define_attribute_user_for_name(node, arg)

        # Caches.

        self.localnames = None # cache for locals

        # Add parameters to the namespace.

        self._add_parameters(argnames)

        # Image generation details.

        self.dynamic = None
        self.location = None
        self.code_location = None
        self.code_body_location = None

        # Program-related details.

        self.blocks = None
        self.body_block = None

        self.temp_usage = 0
        self.local_usage = 0
        self.all_local_usage = 0

    def _add_parameters(self, argnames):

        "Add 'argnames' to the namespace."

        for name in argnames:
            self.set(name, make_instance())

        for name, top_level in self._flattened_parameters(argnames):
            if not top_level:
                self.set(name, make_instance())

    def _flattened_parameters(self, argnames, top_level=1):
        l = []
        for name in argnames:
            if isinstance(name, tuple):
                l += self._flattened_parameters(name, 0)
            else:
                l.append((name, top_level))
        return l

    def __repr__(self):
        if self.location is not None:
            return "<function %s (at %r, code at %r)>" % (
                shortrepr(self), self.location, self.code_location
                )
        else:
            return "<function %s>" % shortrepr(self)

    def __shortrepr__(self):
        return "%s.%s(%s)" % (shortrepr(self.parent), self.name, ", ".join(self.argnames))

    def get_body_block(self):
        return self.body_block

    def is_lambda(self):
        return self._is_lambda

    # Defaults-related methods.

    def store_default(self, attr_or_value):

        """
        Reserve space for defaults, set outside the function, potentially on a
        dynamic basis, using the 'attr_or_value'.
        """

        attr = Attr(None, self, None)
        self._set_using_attr(attr, attr_or_value)
        self.default_attrs.append(attr)

    def make_dynamic(self):

        "Return whether this function must be handled using a dynamic object."

        if self.dynamic is None:
            for attr in self.default_attrs:
                if not attr.is_strict_constant() and self.dynamic_def:
                    self.dynamic = 1
                    self._make_dynamic()
                    break
            else:
                self.dynamic = 0

        return self.dynamic

    is_dynamic = make_dynamic

    def _make_dynamic(self):

        "Where functions have dynamic defaults, add a context argument."

        name = "<context>"
        self.argnames.insert(0, name)
        self.positional_names.insert(0, name)
        self.set(name, make_instance())

    # Namespace-related methods.

    def make_global(self, name):

        "Declare 'name' as a global in the current namespace."

        if name not in self.argnames and not self.has_key(name):
            self.globals.add(name)
            return 1
        else:
            return 0

    def parameters(self):

        """
        Return a dictionary mapping parameter names to their position in the
        parameter list.
        """

        parameters = {}
        for i, name in enumerate(self.argnames):
            parameters[name] = i
        return parameters

    def tuple_parameters(self, argnames=None):

        """
        Return a list of (position, parameter) entries corresponding to tuple
        parameters, where each parameter may either be a string or another such
        list of entries.
        """

        names = argnames or self.argnames

        l = []
        for i, name in enumerate(names):
            if isinstance(name, tuple):
                l.append((i, self.tuple_parameters(name)))
            elif argnames:
                l.append((i, name))
        return l

    def all_locals(self):

        "Return a dictionary mapping names to local and parameter details."

        return dict(self)

    def locals(self):

        "Return a dictionary mapping names to local details."

        if self.localnames is None:
            self.localnames = {}
            self.localnames.update(self.all_locals())
            for name in self.argnames:
                del self.localnames[name]
        return self.localnames

    def is_method(self):

        """
        Return whether this function is a method explicitly defined in a class.
        """

        return isinstance(self.parent, Class)

    def is_relocated(self, name):

        """
        Determine whether the given attribute 'name' is relocated for instances
        having this function as a method.
        """

        for cls in self.parent.descendants:
            if name in cls.relocated:
                return 1
        return 0

    # Administrative methods.

    def items_for_vacuum(self):
        return self.lambdas.items()

    def vacuum_item(self, name):
        del self.lambdas[name]
        return 1

    def finalise_attributes(self):

        """
        Make sure all attributes (locals) are fully defined. Note that locals
        are not attributes in the sense of class, module or instance attributes.
        Defaults are also finalised by this method.
        """

        if self.finalised:
            return

        # Defaults.

        for i, default in enumerate(self.default_attrs):
            default.position = i

        # Parameters.

        i = self._finalise_parameters()

        if i is not None:
            nparams = i + 1
        else:
            nparams = 0

        # Locals (and tuple parameter names).

        i = None
        for i, attr in enumerate(self.locals().values()):
            attr.position = i + nparams

        if i is not None:
            nothers = i + 1
        else:
            nothers = 0

        self.local_usage = nothers
        self.all_local_usage = nparams + nothers
        self.finalised = 1

    def _finalise_parameters(self):
        if not self.argnames:
            return None

        for i, name in enumerate(self.argnames):
            self[name].position = i

        return i

    def as_instantiator(self):

        "Make an instantiator function from a method, keeping all arguments."

        function = Function(self.parent.name, self.parent.parent, self.argnames, self.defaults,
            self.has_star, self.has_dstar, self.dynamic_def, self.module)
        function.default_attrs = self.default_attrs
        return function

class UnresolvedName(NamespaceDict, Constant):

    "A module, class or function which was mentioned but could not be imported."

    def __init__(self, name, parent_name, module=None):
        NamespaceDict.__init__(self, module)
        self.name = name
        self.parent_name = parent_name
        self.parent = None

        self.descendants = set()

    def add_descendant(self, cls):
        self.descendants.add(cls)

    def all_attributes(self):
        return {}

    def all_attribute_names(self):
        return []

    all_class_attributes = class_attributes = instance_attributes = all_attributes
    all_class_attribute_names = class_attribute_names = instance_attribute_names = all_attribute_names

    def __repr__(self):
        return "<unknown %s>" % shortrepr(self)

    def __shortrepr__(self):
        return "%s.%s" % (self.parent_name, self.name)

    def full_name(self):
        if self.name is not None:
            return self.parent_name + "." + self.name
        else:
            return self.parent_name

class Module(NamespaceDict, Constant):

    "An inspected module's core details."

    def __init__(self, name, importer):
        NamespaceDict.__init__(self, self)
        self.name = name
        self.importer = importer
        self.parent = None

        # Original location details.

        self.astnode = None

        # Complete lists of classes and functions.

        self.all_objects = set()

        # Keyword records.

        self.keyword_names = set()

        # Image generation details.

        self.location = None
        self.code_location = None

        # Program-related details.

        self.blocks = None
        self.temp_usage = 0
        self.local_usage = 0
        self.all_local_usage = 0

    def full_name(self):
        return self.name

    def __repr__(self):
        if self.location is not None:
            return "<module %s (at %r)>" % (self.name, self.location)
        else:
            return "<module %s>" % shortrepr(self)

    def __shortrepr__(self):
        return self.name

    # Attribute methods.

    "Return the module attribute names provided by the module."

    module_attribute_names = NamespaceDict.keys

    def module_attributes(self):

        "Return a dictionary mapping names to module attributes."

        return dict(self)

# Pre-made instances.

type_class = TypeClass("type") # details to be filled in later

# Class construction.

def get_class(name, parent, module, node):

    """
    Return a Class instance for the class with the given 'name', 'parent',
    'module' and 'node'.
    """

    if name == "type" and module.full_name() == "__builtins__":
        type_class.set_context(parent, module, node)
        return type_class
    else:
        return CommonClass(name, parent, module, node)

# vim: tabstop=4 expandtab shiftwidth=4
